//+---------------------------------------------------------------------------------+
//| High-Low Cloud Trend [ChartPrime]                                               |
//|                                                                                 |
//| This script is based on the provided Pine Script and is                         |
//| subject to the terms of the Mozilla Public License 2.0                          |
//| © ChartPrime                                                                    |
//+---------------------------------------------------------------------------------+
//| https://www.tradingview.com/script/1yp5FQR8-High-Low-Cloud-Trend-ChartPrime/    |
//+---------------------------------------------------------------------------------+

/*

// This Pine Script™ code is subject to the terms of the Mozilla Public License 2.0 at https://mozilla.org/MPL/2.0/
// © ChartPrime

//@version=5
indicator("High-Low Cloud Trend [ChartPrime]", overlay = true, max_labels_count = 500)


// --------------------------------------------------------------------------------------------------------------------}
// ???????? ????????????
// --------------------------------------------------------------------------------------------------------------------{
len = input.int(35, title="Lookback Period") 

bool mean_rev = input.bool(true, "Mean Reversion Points")
bool show_p_v = input.bool(false, "Volume/Price")

// Colors for trend identification
color_up = color.aqua  
color_dn = color.orange  


// --------------------------------------------------------------------------------------------------------------------}
// ?????????????????? ????????????????????????
// --------------------------------------------------------------------------------------------------------------------{
// Draw Label Function
drawLabel(_pivot, _style, _color, _textColor, h_l) =>
    src = (h_l ? high[1] : low[1])
    if _pivot
        label.new(bar_index[1], src, str.tostring(src, show_p_v ? format.volume : format.mintick), 
                 style =_style, color =_color, textcolor =_textColor, size = size.small)


// Calculate the highest and lowest over the lookback period
highest  = ta.highest(len)  // Highest high over the lookback period
lowest   = ta.lowest(len)  // Lowest low over the lookback period

// Calculate the highest and lowest over a shorter period (1/3 of the original period)
highest1 = ta.highest(len / 4)  // Highest high over the shortened lookback period
lowest1  = ta.lowest(len / 4)  // Lowest low over the shortened lookback period

// Initialize the 'value' variable using 'var' to retain its value across bars
var float value = na

// Logic to update the 'value' variable based on certain conditions
switch
    low  == lowest  => value := highest  // Set 'value' to 'highest' when the current low equals the lowest
    high == highest => value := lowest  // Set 'value' to 'lowest' when the current high equals the highest
    => value := nz(value[1])           // Retain the previous 'value' if neither condition is met

// Determine the 'band' value based on the relationship between 'close' and 'value'
band  = close > value ? lowest : highest  
band1 = close > value ? lowest1 : highest1  

// Determine the color based on the relationship between 'close' and 'value'
color = close > value ? color_up : color_dn  // Use 'color_up' for upward trends and 'color_dn' for downward trends



// --------------------------------------------------------------------------------------------------------------------}
// ??????????????????????????
// --------------------------------------------------------------------------------------------------------------------{

// Plot the 'band' value with a line break if there's a cross between 'close' and 'band'
p1 = plot(ta.cross(close, band) ? na : band, 
              title     = "Band",         
              color     = color,          
              LineWidth = 1,               
              style     = plot.style_linebr)  

// Plot 'band1', which is based on the shorter lookback period, with a transparent color
p2 = plot(band1, 
             title      = "Band1",           
             color      = color.new(chart.fg_color, 90),  
             LineWidth  = 1,                  
             style      = plot.style_linebr)  

// Plot a diamond shape at the 'band' value when there's a cross between 'close' and 'band'
plotshape(ta.cross(close, band) ? band : na,
             location = location.absolute,  
             style    = shape.diamond,      
             color    = color,              
             size     = size.tiny,         
             offset   = 1)                 


// Fill the area between 'p1' and 'p2' (between 'band' and 'band1') using the gradient color
fill(p1, p2, band1, band, color.new(color, 95), color.new(color, 50), fillgaps = true)

// Mean Reversion Conditions
mean_rev_dn = ta.crossunder(high, band1[1]) and band1[1] == high[1] and high != band1 and not ta.cross(close, band) and band[1] == band
mean_rev_up = ta.crossover(low, band1[1]) and band1[1] == low[1]  and low  != band1 and not ta.cross(close, band) and band[1] == band


if mean_rev
    drawLabel(mean_rev_dn, label.style_label_down, color.new(color, 100), chart.fg_color, true)
    drawLabel(mean_rev_up, label.style_label_up, color.new(color, 100), chart.fg_color, false)


// Plot downward arrows above the bars if the previous 'band1' is equal to the previous high and the current high is not equal to 'band1'
plotchar(mean_rev_dn and mean_rev? high[1] : na, 
         "", "?",                    
         location = location.absolute, 
         color    = color,              
         size     = size.tiny,           
         offset   = -1)
//
// Plot upward arrows below the bars if the previous 'band1' is equal to the previous low and the current low is not equal to 'band1'
plotchar(mean_rev_up and mean_rev? low[1] : na, 
         "", "?",                       
         location = location.absolute,   
         color    = color,             
         size     = size.tiny,
         offset   = -1)         

// --------------------------------------------------------------------------------------------------------------------}


*/




#property indicator_chart_window
#property indicator_buffers 13
#property indicator_plots   7

// Indicator buffers
double BandBuffer[];
double Band1Buffer[];
double BandBufferUP[];
double Band1BufferUP[];
double BandBufferDN[];
double Band1BufferDN[];
double UpArrowBuffer[];
double DownArrowBuffer[];
double HighestBuffer[];
double LowestBuffer[];
double Highest1Buffer[];
double Lowest1Buffer[];
double ValueBuffer[];

// Input parameters
extern int             LookbackPeriod  = 36;
extern bool            MeanRevPoints   = true;
extern bool            ShowVolumePrice = false;
extern bool            BandFilled      = true;

extern string HistoSettings1;
input bool             UseAutoWidth    = true;           // Auto adjust candle body width
extern int             HistoWidth      = 4;              // Histo width
// Colors for trend identification
extern color           ColorUp         = clrDodgerBlue;
extern color           ColorDn         = clrIndianRed;
extern int             LineWidth       = 2;
extern color           LineColorUp     = clrSpringGreen;
extern color           LineColorDn     = clrMagenta;
extern int             ArrowWidth      = 2;
extern color           ArrowColorUp    = clrAqua;
extern color           ArrowColorDn    = clrGold;
extern int             ArrowCode       = 116;

extern bool            alertsOn        = true;                // Turn alerts on?
extern bool            alertsOnCurrent = false;               // Alerts on still opened bar?
extern bool            alertsMessage   = true;                // Alerts should display message?
extern bool            alertsSound     = true;                // Alerts should play a sound?
extern bool            alertsNotify    = false;               // Alerts should send a notification?
extern bool            alertsEmail     = false;               // Alerts should send an email?
extern string          soundFile       = "alert2.wav";        // Alerts sound file

 
extern int                btn_Subwindow         = 1;
extern ENUM_BASE_CORNER   btn_corner            = CORNER_LEFT_UPPER; 
extern string             btn_text              = "HLCT ";
extern string             btn_Font              = "Arial";
extern int                btn_FontSize          = 10;                            
extern color              btn_text_ON_color     = clrLime;
extern color              btn_text_OFF_color    = clrRed;
extern string             btn_pressed           = "HLC on";            
extern string             btn_unpressed         = "HLC off";
extern color              btn_background_color  = clrDimGray;
extern color              btn_border_color      = clrBlack;
extern int                button_x              = 1;                                 
extern int                button_y              = 1;                                   
extern int                btn_Width             = 60;                                 
extern int                btn_Height            = 20;                                
extern string             soundBT               = "tick.wav";  




int    candlewidth = 0;


bool                      show_data             = true;
string IndicatorName, IndicatorObjPrefix ,buttonId ;

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string GenerateIndicatorName(const string target)
{
   string name = target;
   int try = 2;
   while (WindowFind(name) != -1)
   {
      name = target + " #" + IntegerToString(try++);
   }
   return name;
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Custom function to draw labels                                   |
//+------------------------------------------------------------------+
void DrawLabel(int Index, bool Pivot, bool Style, color Color, color TextColor, bool HighLow) {
    double Src = HighLow ? High[Index] : Low[Index];
    if (Pivot) {
        string Text = ShowVolumePrice ? DoubleToString(Volume[Index], _Digits) : DoubleToString(Src, _Digits);
        string Label = "Label" + IntegerToString(Index);
        ObjectCreate(Label, OBJ_LABEL, 0, Time[Index], Src);
        ObjectSetText(Label, Text, 8, "Arial", TextColor);
        ObjectSet(Label, OBJPROP_COLOR, Color);
        ObjectSet(Label, OBJPROP_BACK, false);
    }
}

//+------------------------------------------------------------------+
//| Custom function to check for mean reversion conditions           |
//+------------------------------------------------------------------+
bool MeanRevDown(int Index) {
    bool temp1 = CrossUnder(Band1Buffer, Index);
    //if (temp1 ) Print("Test MRDN: " + (string) temp1 + "@ " + Index);
    return CrossUnder(Band1Buffer, Index) && Band1Buffer[Index + 1] == High[Index + 1] && High[Index] != Band1Buffer[Index] && !Cross(BandBuffer, Index) && BandBuffer[Index] == BandBuffer[Index + 1];
}

bool MeanRevUp(int Index) {
    bool temp1 = CrossOver(Band1Buffer, Index) ;
    //if (temp1 ) Print("Test MRUP: " + (string) temp1 + "@ " + Index);
    return CrossOver(Band1Buffer, Index) && Band1Buffer[Index + 1] == Low[Index + 1] && Low[Index] != Band1Buffer[Index] && !Cross(BandBuffer, Index) && BandBuffer[Index] == BandBuffer[Index + 1];
}

//+------------------------------------------------------------------+
//| Custom function to check for cross                               |
//+------------------------------------------------------------------+
bool Cross(double &Line2[], int Index) {
    return (Close[Index + 1] <= Line2[Index + 1] && Close[Index] > Line2[Index]) || (Close[Index + 1] >= Line2[Index + 1] && Close[Index] < Line2[Index]);
}

bool CrossOver(double &Line2[], int Index) {
    return Low[Index + 1] <= Line2[Index + 1] && Low[Index] > Line2[Index];
}

bool CrossUnder(double &Line2[], int Index) {
    return High[Index + 1] >= Line2[Index + 1] && High[Index] < Line2[Index];
}

//+------------------------------------------------------------------+
//| Indicator initialization function                                |
//+------------------------------------------------------------------+
int OnInit() {


   if (UseAutoWidth)
   {
      int scale = int(ChartGetInteger(0,CHART_SCALE));
      switch(scale) 
	   {
	      case 0: candlewidth =  1; break;
	      case 1: candlewidth =  1; break;
		   case 2: candlewidth =  2; break;
		   case 3: candlewidth =  3; break;
		   case 4: candlewidth =  6; break;
		   case 5: candlewidth = 14; break;
	   }
	}
	else { candlewidth = HistoWidth; }
	
	
    // Set indicator buffers
    SetIndexBuffer(0, BandBuffer);
    SetIndexBuffer(1, Band1Buffer);
    SetIndexBuffer(2, BandBufferUP);
    SetIndexBuffer(3, Band1BufferUP);
    SetIndexBuffer(4, BandBufferDN);
    SetIndexBuffer(5, Band1BufferDN);
    SetIndexBuffer(6, UpArrowBuffer);
    SetIndexBuffer(7, DownArrowBuffer);
    SetIndexBuffer(8, HighestBuffer);
    SetIndexBuffer(9, LowestBuffer);
    SetIndexBuffer(10, Highest1Buffer);
    SetIndexBuffer(11, Lowest1Buffer);
    SetIndexBuffer(12, ValueBuffer);

    SetIndexLabel(0,"Outer Band");
	 SetIndexLabel(1,"Inner Band");
    SetIndexLabel(2,"Outer Band Up");
	 SetIndexLabel(3,"Inner Band Up");
    SetIndexLabel(4,"Outer Band Down");
	 SetIndexLabel(5,"Inner Band Down");
	 SetIndexLabel(6,"Mean Reversion Up");
	 SetIndexLabel(7,"Mean Reversion Down");
	 SetIndexLabel(8,"Outer Highest Channel");
	 SetIndexLabel(9,"Inner Highest Channel");
	 SetIndexLabel(10,"Outer Lowest Channel");
	 SetIndexLabel(11,"Inner Lowest Channel");
	 SetIndexLabel(12,"Direction Value");

    // Set plot styles
    if(BandFilled)
    {
       SetIndexStyle(0, DRAW_HISTOGRAM, STYLE_SOLID, candlewidth, ColorDn);
       SetIndexStyle(1, DRAW_HISTOGRAM, STYLE_SOLID, candlewidth, ColorUp);
       SetIndexStyle(2, DRAW_NONE);
       SetIndexStyle(3, DRAW_NONE);
       SetIndexStyle(4, DRAW_NONE);
       SetIndexStyle(5, DRAW_NONE);
    }else
    {
       SetIndexStyle(0, DRAW_NONE);
       SetIndexStyle(1, DRAW_NONE);
       SetIndexStyle(2, DRAW_LINE, 0, LineWidth, LineColorUp);
       SetIndexStyle(3, DRAW_LINE, 0, LineWidth, LineColorUp);
       SetIndexStyle(4, DRAW_LINE, 0, LineWidth, LineColorDn);
       SetIndexStyle(5, DRAW_LINE, 0, LineWidth, LineColorDn);
    }
    SetIndexStyle(6,DRAW_ARROW,STYLE_SOLID,ArrowWidth,ArrowColorUp);
    SetIndexArrow(6,ArrowCode);
    SetIndexStyle(7,DRAW_ARROW,STYLE_SOLID,ArrowWidth,ArrowColorDn);
    SetIndexArrow(7,ArrowCode);	
    SetIndexStyle(8, DRAW_NONE);
    SetIndexStyle(9, DRAW_NONE);
    SetIndexStyle(10, DRAW_NONE);
    SetIndexStyle(11, DRAW_NONE);
	 SetIndexStyle(12, DRAW_NONE);
    
	IndicatorName = GenerateIndicatorName(btn_text);
   IndicatorObjPrefix = "__" + IndicatorName + "__";
   
   
      double valz;
   if (GlobalVariableGet(IndicatorName + "_visibility", valz))
   show_data = valz != 0;

   ChartSetInteger(ChartID(), CHART_EVENT_MOUSE_MOVE, 1);
   buttonId = IndicatorObjPrefix+btn_text;
   createButton(buttonId, btn_text, btn_Width, btn_Height, btn_Font, btn_FontSize, btn_background_color, btn_border_color, btn_text_ON_color);
   ObjectSetInteger(ChartID(), buttonId, OBJPROP_YDISTANCE, button_y);
   ObjectSetInteger(ChartID(), buttonId, OBJPROP_XDISTANCE, button_x); 

    return(INIT_SUCCEEDED);
}



void createButton(string buttonID,string buttonText,int width,int height,string font,int fontSize,color bgColor,color borderColor,color txtColor)
{
      ObjectDelete    (ChartID(),buttonID);
      ObjectCreate(ChartID(),buttonID,OBJ_BUTTON,btn_Subwindow,0,0);
      ObjectSetInteger(ChartID(),buttonID,OBJPROP_COLOR,txtColor);
      ObjectSetInteger(ChartID(),buttonID,OBJPROP_BGCOLOR,bgColor);
      ObjectSetInteger(ChartID(),buttonID,OBJPROP_BORDER_COLOR,borderColor);
      ObjectSetInteger(ChartID(),buttonID,OBJPROP_XSIZE,width);
      ObjectSetInteger(ChartID(),buttonID,OBJPROP_YSIZE,height);
      ObjectSetString (ChartID(),buttonID,OBJPROP_FONT,font);
      ObjectSetString (ChartID(),buttonID,OBJPROP_TEXT,buttonText);
      ObjectSetInteger(ChartID(),buttonID,OBJPROP_FONTSIZE,fontSize);
      ObjectSetInteger(ChartID(),buttonID,OBJPROP_SELECTABLE,0);
      ObjectSetInteger(ChartID(),buttonID,OBJPROP_CORNER,btn_corner);
      ObjectSetInteger(ChartID(),buttonID,OBJPROP_HIDDEN,1);
      ObjectSetInteger(ChartID(),buttonID,OBJPROP_XDISTANCE,9999);
      ObjectSetInteger(ChartID(),buttonID,OBJPROP_YDISTANCE,9999);
}


//+------------------------------------------------------------------+
//| Indicator deinitialization function                              |
//+------------------------------------------------------------------+
void OnDeinit(const int reason) 


{
      ObjectsDeleteAll(0,"HLCT ");
      ObjectsDeleteAll(ChartID(), IndicatorObjPrefix);

    // Clean up objects
    for (int i = ObjectsTotal() - 1; i >= 0; i--) {
        string ObjectName = ObjectName(i);
        if (ObjectName != NULL && StringFind(ObjectName, "Label") == 0) 
        
        {
            ObjectDelete(ObjectName);
        
        }
    }
}

bool recalc = true;
 void handleButtonClicks()
{
   if (ObjectGetInteger(ChartID(), buttonId, OBJPROP_STATE))
   {
      ObjectSetInteger(ChartID(), buttonId, OBJPROP_STATE, false);
      show_data = !show_data;
      GlobalVariableSet(IndicatorName + "_visibility", show_data ? 1.0 : 0.0);
      recalc = true;
      start();
   }
}
void OnChartEvent(const int id, 
                  const long &lparam,
                  const double &dparam,
                  const string &sparam)
{
   handleButtonClicks();
   if (id==CHARTEVENT_OBJECT_CLICK && ObjectGet(sparam,OBJPROP_TYPE)==OBJ_BUTTON)
   {
   if (soundBT!="") PlaySound(soundBT);     
   }
   
  
	
}
//
//
//
//

int start()
{

handleButtonClicks();
   recalc = false;
   
   SetIndexStyle(0,DRAW_HISTOGRAM);
   SetIndexStyle(1,DRAW_HISTOGRAM);
   SetIndexStyle(2,DRAW_LINE);
   SetIndexStyle(3,DRAW_LINE);
   SetIndexStyle(4,DRAW_LINE);
   SetIndexStyle(5,DRAW_LINE);
   SetIndexStyle(6,DRAW_ARROW);
   SetIndexStyle(7,DRAW_ARROW);
  
   start2();
   
  
   
      
   
   if (show_data)
      {
      ObjectSetInteger(ChartID(),buttonId,OBJPROP_COLOR,btn_text_ON_color);
      ObjectSetString(ChartID(),buttonId,OBJPROP_TEXT,btn_unpressed);
     
      }
      else
      {
      ObjectSetInteger(ChartID(),buttonId,OBJPROP_COLOR,btn_text_OFF_color);
      ObjectSetString(ChartID(),buttonId,OBJPROP_TEXT,btn_pressed);
      
   SetIndexStyle(0,DRAW_NONE);
   SetIndexStyle(1,DRAW_NONE);
   SetIndexStyle(2,DRAW_NONE);
   SetIndexStyle(3,DRAW_NONE);
   SetIndexStyle(4,DRAW_NONE);
   SetIndexStyle(5,DRAW_NONE);
   SetIndexStyle(6,DRAW_NONE);
   SetIndexStyle(7,DRAW_NONE);      
   
      
   
      //template code     
      }
       return(0);
      }



//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
//
//
//

//+------------------------------------------------------------------+
//| Indicator iteration function                                     |
//+------------------------------------------------------------------+
int start2()
{ 
int counted_bars=IndicatorCounted();
   
   
   if(counted_bars < 0) return(-1);
   if(counted_bars > 0) counted_bars--;
                          
 

    int limit = Bars - counted_bars;
    if (counted_bars > 1) limit++;
    
    
    if(ChartGetInteger(0,CHART_SCALE) != candlewidth) OnInit();

    for (int i = limit -1 ; i >= 0; i--) {
        // Calculate highest and lowest over the lookback period
        double Highest = High[iHighest(Symbol(), Period(), MODE_HIGH, LookbackPeriod, i)];
        double Lowest = Low[iLowest(Symbol(), Period(), MODE_LOW, LookbackPeriod, i)];

        // Calculate highest and lowest over a shorter period (1/3 of the original period)
        double Highest1 = High[iHighest(Symbol(), Period(), MODE_HIGH, (int) LookbackPeriod / 4, i)];
        double Lowest1 = Low[iLowest(Symbol(), Period(), MODE_LOW, (int) LookbackPeriod / 4, i)];

        // Initialize the 'value' variable
        static double Value = 0.0;

        // Logic to update the 'value' variable based on certain conditions
        if (Low[i] == Lowest) {
            Value = Highest;
        } else if (High[i] == Highest) {
            Value = Lowest;
        } else {
            Value = ValueBuffer[i+1];
        }

        // Determine the 'band' value based on the relationship between 'close' and 'value'
        if(Close[i] > Value)
        {
           BandBufferUP[i] = Lowest;
           Band1BufferUP[i] = Lowest1;
           BandBuffer[i] = Lowest;
           Band1Buffer[i] = Lowest1;
        }else
        {
           BandBufferDN[i] = Highest;
           Band1BufferDN[i] = Highest1;
           BandBuffer[i] = Highest;
           Band1Buffer[i] = Highest1;          
        }
        
        double Band = Close[i] > Value ? Lowest : Highest;
        double Band1 = Close[i] > Value ? Lowest1 : Highest1;

        // Determine the color based on the relationship between 'close' and 'value'
        color Color = Close[i] > Value ? ColorUp : ColorDn;
        
        // Store the calculated values in the buffers        
        HighestBuffer[i] = Highest;
        LowestBuffer[i] = Lowest;
        Highest1Buffer[i] = Highest1;
        Lowest1Buffer[i] = Lowest1;
        ValueBuffer[i] = Value;

        // Plot upward and downward arrows for mean reversion points
        if (MeanRevPoints) {
            if (MeanRevUp(i)) {
                UpArrowBuffer[i] = Low[i];
                //Print("Got in MeanRevUp");
                //DrawLabel(i, true, true, Color, clrWhite, false);
            } else {
                UpArrowBuffer[i] = EMPTY_VALUE;
            }
            

            if (MeanRevDown(i)) {
                DownArrowBuffer[i] = High[i];
                //Print("Got in MeanRevDn");
                //DrawLabel(i, true, false, Color, clrWhite, true);
            } else {
                DownArrowBuffer[i] = EMPTY_VALUE;
            }
        } else {
            UpArrowBuffer[i] = EMPTY_VALUE;
            DownArrowBuffer[i] = EMPTY_VALUE;
        }
    }

    if (alertsOn)
    {
      int whichBar = 1; if (alertsOnCurrent) whichBar = 0;
      if ((BandBufferUP[whichBar+1] == EMPTY_VALUE || BandBufferUP[whichBar+1] == 0) && (BandBufferUP[whichBar] != EMPTY_VALUE && BandBufferUP[whichBar] != 0)) doAlert(" UPWARD Trend ");
      if ((BandBufferDN[whichBar+1] == EMPTY_VALUE || BandBufferDN[whichBar+1] == 0) && (BandBufferDN[whichBar] != EMPTY_VALUE && BandBufferDN[whichBar] != 0)) doAlert(" DOWNWARD Trend ");
      if (UpArrowBuffer[whichBar] != 0 && UpArrowBuffer[whichBar] != EMPTY_VALUE)  doAlert("Mean Reversion UP"); 
	   if (DownArrowBuffer[whichBar] != 0 && DownArrowBuffer[whichBar] != EMPTY_VALUE)  doAlert("Mean Reversion DOWN"); 
    } 

    return(0);
}



//------------------------------------------------------------------
//                                                                  
//------------------------------------------------------------------

void doAlert(string doWhat)
{
   static string   previousAlert="nothing";
   static datetime previousTime;
   string message;
   
      if (previousAlert != doWhat || previousTime != Time[0]) {
          previousAlert  = doWhat;
          previousTime   = Time[0];

          //
          //

          message = timeFrameToString(_Period)+" "+_Symbol+" at "+TimeToStr(TimeLocal(),TIME_SECONDS)+" High Low Cloud Trend - "+doWhat;
             if (alertsMessage) Alert(message);
             if (alertsNotify)  SendNotification(message);
             if (alertsEmail)   SendMail(_Symbol+" High Low Cloud Trend - ",message);
             if (alertsSound)   PlaySound(soundFile);
      }
}

//
//
//

string sTfTable[] = {"M1","M5","M15","M30","H1","H4","D1","W1","MN"};
int    iTfTable[] = {1,5,15,30,60,240,1440,10080,43200};

string timeFrameToString(int tf)
{
   for (int i=ArraySize(iTfTable)-1; i>=0; i--) 
         if (tf==iTfTable[i]) return(sTfTable[i]);
                              return("");
}